const { ExtensionCommon } = ChromeUtils.import("resource://gre/modules/ExtensionCommon.jsm");

const { Services } = ChromeUtils.import("resource://gre/modules/Services.jsm");
const { AddonManager } = ChromeUtils.import("resource://gre/modules/AddonManager.jsm");

const { MailServices }=ChromeUtils.import("resource:///modules/MailServices.jsm");
const { MailUtils }=ChromeUtils.import("resource:///modules/MailUtils.jsm");
const { FileUtils }=ChromeUtils.import("resource://gre/modules/FileUtils.jsm");


const EXTENSION_ID = 'simplemailredirection@ggbs.de';

const MAX_HEADER_LENGTH = 16384;

var prefs={'debug': true};
var windows=new Object;

//debug('entered');
var smr = class extends ExtensionCommon.ExtensionAPI {
  onStartup() { 
//pres.debug not set yet!
//debug('onStartup');
  }

  onShutdown(isAppShutdown) {
debug('onShutdown isAppShutdown='+isAppShutdown);
      if (isAppShutdown) return;
      // Looks like we got uninstalled. Maybe a new version will be installed
      // now. Due to new versions not taking effect
      // (https://bugzilla.mozilla.org/show_bug.cgi?id=1634348)
      // we invalidate the startup cache. That's the same effect as starting
      // with -purgecaches (or deleting the startupCache directory from the
      // profile).
//TODO: remove stylesheet
      Services.obs.notifyObservers(null, "startupcache-invalidate");
  }
  getAPI(context) {
//debug('getApi entered');	//more than once!

    context.callOnClose(this);
		this.getAddon();


    return {
      smr: {
        init: async function(options) {
          prefs=options;
debug('init debug='+prefs.debug);
					let m3p=Services.wm.getMostRecentWindow("mail:3pane");
          // If loaded immediately, there is a sideeffect which prevets cs2c from running
          // The error is:
          m3p.setTimeout(stylesheets, 1000, context);
				},
        redirect: async function(mhs, rts, windowId, options) {
          prefs=options;
debug('redirect from window '+windowId);
mhs.forEach(mh=>debug('redirect '+mh.subject+' ('+mh.author+')'));
rts.forEach(add=>debug('resentTo '+add));
          windows[windowId].msgCount=mhs.length;
          mhs.forEach(mh=>{
            windows[windowId][mh.id]=mh;
            windows[windowId][mh.id].state='started';
          });
//TEST:
//mhs.forEach(mh=>windows[windowId][mh.id]['msgSend']='send_'+mh.id);
          let resentTo='';
          let sep='';
          rts.forEach(add=>{resentTo+=sep+add; sep=',';});
debug('resendTo='+resentTo);

          mhs.forEach((mh) => {
            let msgHdr=context.extension.messageManager.get(mh.id);
debug('got msgHdr id='+msgHdr.messageKey);
            let accountId=mh.folder.accountId;
            let path=mh.folder.path;
debug('resent '+mh.subject+' ('+mh.author+') to: '+accountId+' '+path+' '+resentTo);
            let account=MailServices.accounts.getAccount(accountId);
            let folderURI=account.incomingServer.rootFolder.URI+path;
            let folder=MailUtils.getExistingFolder(folderURI);
//debug('current keywords: '+msgHdr.getStringProperty("keywords"));
            if (/(?:^| )redirected(?: |$)/.test(msgHdr.getStringProperty("keywords"))) {
debug('..already resent!');
/*
debug('Test: remove redirected indicator from old version');
let msg = Cc["@mozilla.org/array;1"].createInstance(Ci.nsIMutableArray);
msg.appendElement(msgHdr, false);
msgHdr.folder.removeKeywordsFromMessages(msg, "resent");
msgHdr.folder.addKeywordsToMessages(msg, "redirected");
*/
            }
            let msgUri = msgHdr.folder.generateMessageURI(msgHdr.messageKey);
            let identity = account.defaultIdentity;
            let sender=MailServices.headerParser.makeMimeHeader([{name: identity.fullName, email: identity.email}], 1);
debug('call resent URI='+msgUri+' sender='+sender);
            let msgCompFields = Cc["@mozilla.org/messengercompose/composefields;1"].
                             createInstance(Ci.nsIMsgCompFields);
            msgCompFields.to=resentTo;  //this converts to quoted printable!
            msgCompFields.from = sender;
            msgCompFields.fcc = "nocopy://";
            msgCompFields.fcc2 = "nocopy://";
            resentMessage(mh.id, windowId, msgUri, accountId, msgCompFields, identity);
          });
				},
        abort: async function(windowId, msgId) {
debug('abort windowId='+windowId+' msgId='+msgId);
          let msg=windows[windowId][msgId];
          if (msg.msgSend) {
debug('already sending, abort')
            msg.msgSend.abort();
          }
          msg.state='abort';  //mark for abort after filecopy
        },
        onMailSent: new ExtensionCommon.EventManager({
          context,
          name: "smr.onMailSent",
          register(fire, windowIdparam) {
debug('onMailSent register windowId='+windowIdparam);
            let windowId=windowIdparam;
            windows[windowId]=new Object;
            windows[windowId].fire=fire;
            return function() {
debug('onMailSent unregister windowId='+windowId);
              if (!windows[windowId]) return;
              windows[windowId].fire=null;
              for (const [msgId, msg] of Object.entries(windows[windowId])) {
                if (isNaN(msgId)) continue; //.fire or other
debug('  abort '+msgId+' '+msg.subject);
                if (msg.state=='started') {
                  if (msg.msgSend) {
debug('    already sending, abort')
                    msg.msgSend.abort();
                  }
                  msg.state='abort';  //mark for abort after filecopy
                }
              }
            };
          },
        }).api(),
			}
    }
  }
  close() {
    // This function is called if the extension is disabled or removed, or Thunderbird closes.
    // Flush all caches
  }
  async getAddon() {
    let addOn=await AddonManager.getAddonByID(EXTENSION_ID);
    let console = Services.console;
    let app = Services.appinfo;
    console.logStringMessage('SimpleMailRedirection: '+addOn.version+' on '+app.name+' '+app.version);
  }

};

////////////////////////////////////////////////////////////////

function nsMsgSendListener(msgId, windowId, msgUri, tmpFile) {
  this.msgId=msgId;
  this.windowId=windowId;
  this.msgUri=msgUri;
  this.tmpFile=tmpFile;
  this.size=tmpFile.fileSize;
}
nsMsgSendListener.prototype = {
  msgId: null,
  windowId: null,
  msgUri: null,
  tmpFile: null,
  size: 0,

  QueryInterface: function(iid) {
    if (iid.equals(Ci.nsIMsgSendListener) ||
        iid.equals(Ci.nsIMsgCopyServiceListener) ||
        iid.equals(Ci.nsISupports)) {
      return this;
    }
    throw Components.results.NS_NOINTERFACE;
  },
  onProgress(msgId, progress, progressMax) {  //not called :-(
debug('onProgress '+progress+' up to  '+progressMax);
  },
  onStartSending(msgId, msgSize) {
    //msgId is null
    //msgSize is always 0 :-(
    if (windows[this.windowId].fire) {
debug('started: notify popup window');
      windows[this.windowId].fire.async({msgid: this.msgId, type: 'started', size: this.size});
    } else {
debug('started: popup window has vanished');
    }
  },
  onStopSending(aMsgID, aStatus, aMsg, returnFileSpec) {
    //aMsgID is empty
//debug('nsMsgSendListener.onStopSending '+aMsgID+', '+aStatus +', '+aMsg+', '+returnFileSpec+' '+this.msgUri );
//nsMsgSendListener.onStopSending null, 2153066799, null, null imap-message://ggbs@mailhost.iwf.ing.tu-bs.de/INBOX/Gitti#3085debug('onStopSending status='+aStatus);
    if (aStatus) {
      this.tmpFile.remove(false);
    } else {
      // mark message as 'resent'
      let messenger = Cc["@mozilla.org/messenger;1"].
                      createInstance(Ci.nsIMessenger);
      let msgService = messenger.messageServiceFromURI(this.msgUri);
      let msgHdr = msgService.messageURIToMsgHdr(this.msgUri);

      let msg = Cc["@mozilla.org/array;1"].createInstance(Ci.nsIMutableArray);
      msg.appendElement(msgHdr, false);
      try {
        msgHdr.folder.addKeywordsToMessages(msg, "redirected");
      } catch(e) {
debug(e);
      }
//debug('keywords set to: '+msgHdr.getStringProperty("keywords"));
    }
    if (windows[this.windowId].fire) {
debug('finished: notify popup window');
      windows[this.windowId].fire.async({msgid: this.msgId, type: 'finished', state: aStatus});
    } else {
debug('finished: popup window has vanished');
    }
    windows[this.windowId][this.msgId].state='finished';
    windows[this.windowId].msgCount--;
for (const [windowId, data] of Object.entries(windows)) {
  for (const [msgId, mh] of Object.entries(data)) {
    if (isNaN(msgId)) debug(`cache ${windowId} ${msgId} ${mh}`);
    else debug(`cache msgId ${windowId} ${msgId} ${mh.id} ${mh.msgSend} ${mh.subject}`);
  }
}
    if (!windows[this.windowId].msgCount) {
debug('delete windowId '+this.windowId);
      delete windows[this.windowId];
    }

  }
}

function resentMessage(msgId, windowId, uri, accountId, msgCompFields, identity) {
debug('resentMessage ' + uri);

  let tmpFile = FileUtils.getFile("TmpD", ['tb_simplemailredirection.tmp']);
  tmpFile.createUnique(tmpFile.NORMAL_FILE_TYPE, parseInt("0600", 8));
  if (tmpFile === null) {
debug('temp localfile is null');
    return;
  } else {
debug('writing message to '+tmpFile.path);
  }

  let messenger = Cc["@mozilla.org/messenger;1"].
                  createInstance(Ci.nsIMessenger);
  let aScriptableInputStream = Cc["@mozilla.org/scriptableinputstream;1"].
                               createInstance(Ci.nsIScriptableInputStream);
  let aFileOutputStream = Cc["@mozilla.org/network/file-output-stream;1"].
                          createInstance(Ci.nsIFileOutputStream);

  let inHeader = true;
  let skipping = false;
  let leftovers = "";
  let buf = "";
  let line = "";

  let aCopyListener = {
    onStartRequest: function(aRequest, aContext) {
      // write out Resent-* headers
      let resenthdrs = getResentHeaders(msgCompFields, identity);
debug('resent with account='+accountId+' identity='+identity.fullName+' <'+identity.email+'> => sender='+msgCompFields.from);
      aFileOutputStream.write(resenthdrs, resenthdrs.length);
    },

    onStopRequest: function(aRequest, aContext, aStatusCode) {
      // write leftovers
      aFileOutputStream.write(leftovers, leftovers.length);
      aFileOutputStream.close();

      if (aStatusCode) {
debug('aCopyListener.onStopRequest failed '+aRequest+', '+aContext+', '+aStatusCode);
        return;
      }
let test=msgCompFields.to.toLowerCase().includes('testggbs@ggbs.de');
      if (windows[windowId][msgId].state=='abort') {
debug('abort: remove tmpfile');
        tmpFile.remove(false);
        if (windows[windowId].fire) {
          windows[windowId].fire.async({msgid: msgId, type: 'aborted', state: 1});
        } else {
debug('copied: popup window has vanished');
        }
        windows[windowId].msgCount--;
        if (!windows[windowId].msgCount) {
debug('delete windowId '+windowId);
          delete windows[windowId];
        }
        return;
      }
//TEST
debug('file copied to tempfile '+tmpFile.path);
if (test) {
console.log('SMR: TEST mode, no msgSend');  //even if no debug
for (const [windowId, data] of Object.entries(windows)) {
  for (const [msgId, mh] of Object.entries(data)) {
    if (isNaN(msgId)) debug(`cache ${windowId} ${msgId} ${mh}`);
    else debug(`cache msgId ${windowId} ${msgId} ${mh.id} ${mh.msgSend} ${mh.subject}`);
  }
}
windows[windowId][msgId].state='finished';
let m3p=Services.wm.getMostRecentWindow("mail:3pane");
m3p.setTimeout(()=>{
  if (windows[windowId].fire)
    windows[windowId].fire.async({msgid: msgId, type: 'finished', status: 1})
  windows[windowId].msgCount--;
  if (!windows[windowId].msgCount) {
  debug('delete windowId '+windowId);
    delete windows[windowId];
  }
}, 3000);
return;
}
debug('actually resent with account='+accountId+' identity='+identity.fullName+' <'+identity.email+'> => sender='+msgCompFields.from);

      // send a message
      let msgSendListener = new nsMsgSendListener(msgId, windowId, uri, tmpFile);
      let msgSend = Cc["@mozilla.org/messengercompose/send;1"].
                    createInstance(Ci.nsIMsgSend);
      windows[windowId][msgId].msgSend=msgSend;
//let progress=msgSend.getProgress(); // is null
//debug('progress='+progress);
      try {
        msgSend.sendMessageFile(
          identity,                        // in nsIMsgIdentity       aUserIdentity,
          accountId,                       // char* accountKey,
          msgCompFields,                   // in nsIMsgCompFields     fields,
          tmpFile,                         // in nsIFile              sendIFile,
          true/*!prefs.debug*/,                            // in PRBool               deleteSendFileOnCompletion,
          false,                           // in PRBool               digest_p,
          msgSend.nsMsgDeliverNow,         // in nsMsgDeliverMode     mode,
          null,                            // in nsIMsgDBHdr          msgToReplace,
          msgSendListener,                 // in nsIMsgSendListener   aListener,
          null,                            // in nsIMsgStatusFeedback aStatusFeedback,
          ""                               // in string               password
          );
      } catch(e) {
        console.log('SMR: exception when sending message:' + e);
      }
    },

    onDataAvailable: function(aRequest, aInputStream, aOffset, aCount) {
//debug('onDataAvailable');
      if (windows[windowId][msgId].state=='abort') {
//debug('onDataAvailable aborted');
//        throw 'aborted';
          //see https://developer.mozilla.org/en-US/docs/Mozilla/Tech/XPCOM/Reference/Interface/nsIStreamListener
          //but does not work
        return;
      }

      aScriptableInputStream.init(aInputStream);
      //dumper.dump("!! inHeader reading new buffer, " + aCount + " bytes");
      buf = leftovers + aScriptableInputStream.read(aCount);
      if (inHeader) {
        leftovers = "";

        while (buf.length > 0) {
          // find end of line
          let eol = buf.indexOf("\r");

          if (eol === -1) {
            // no end of line character in buffer
            // remember this part for the next time
            leftovers = buf;
            // dumper.dump("leftovers=>>"+leftovers+"<<leftovers_end. length=" + leftovers.length);
            break;
          } else {
            // try a pair of eol chars
            if (eol + 1 < buf.length) {
              eol++;  // forward to \n
            } else {
              // pair couldn't be found because of end of buffer
              // dumper.dump("pair couldn't be found. end of buf. eol="+eol+" buf.length="+buf.length);
              leftovers = buf;
              break;
            }
            line = buf.substr(0, eol+1);
            buf = buf.substr(eol+1);
            // dumper.dump("line=>>"+line+"<<line_end. length=" + line.length);

            if (line === "\r\n") {
              aFileOutputStream.write(line, line.length);
              inHeader = false;
              leftovers = buf;
              break;
            }
          }

          if (skipping) {
            if (line[0] === " " || line[0] === "\t") {
              // dumper.dump("forbidden line:" + line+"<<");
              // continue;
            } else {
              skipping = false;
            }
          }

/*web.de
          if (/^from: /i.test(line)) {
            line="From: "+msgCompFields.from+"\r\n";
debug('rewrite From: '+msgCompFields.from);
          }
*/
          // remove sensitive headers (vide: nsMsgSendPart.cpp)
          // From_ line format - http://www.qmail.org/man/man5/mbox.html
          if (/^[>]*From \S+ /.test(line) ||
              /^bcc: /i.test(line) ||
              /^resent-bcc: /i.test(line) ||
              /^fcc: /i.test(line) ||
              /^content-length: /i.test(line) ||
              /^lines: /i.test(line) ||
              /^status: /i.test(line) ||
              /^x-mozilla-status(?:2)?: /i.test(line) ||
              /^x-mozilla-draft-info: /i.test(line) ||
              /^x-mozilla-newshost: /i.test(line) ||
              /^x-uidl: /i.test(line) ||
              /^x-vm-\S+: /i.test(line) ||
              /^return-path: /i.test(line) ||
              /^delivered-to: /i.test(line) ||

              // for drafts
              /^FCC: /i.test(line) ||
              /^x-identity-key: /i.test(line) ||
              /^x-account-key: /i.test(line)) {
            skipping = true;            // discard line
debug('removed line: ' + line);
          }

          if (!skipping) {
            var ret = aFileOutputStream.write(line, line.length);
            if (ret !== line.length) {
              debug("!! inHeader write error? line len "+ line.length + ", written "+ ret);
            }
          }
        }

        if (!inHeader && leftovers !== "") {
          // convert all possible line terminations to CRLF (required by RFC822)
//          leftovers = leftovers.replace(/\r\n|\n\r|\r|\n/g, "\r\n");
          ret = aFileOutputStream.write(leftovers, leftovers.length);
          if (ret !== leftovers.length) {
            dumper.dump("!! inBody write error? leftovers len " + leftovers.length + ", written " + ret);
          }
          leftovers = "";
        }
      } else {
        // out of header -- read the rest and write to file
        leftovers = "";
        // convert all possible line terminations to CRLF (required by RFC822)
//        buf = buf.replace(/\r\n|\n\r|\r|\n/g, "\r\n");
        ret = aFileOutputStream.write(buf, buf.length);
        if (ret !== buf.length) {
          dumper.dump("!! inBody write error? buf len " + buf.length + ", written " + ret);
        }
        buf = "";
      }
    } //End of onDataAvailable
  } // End of aCopyListener

  var msgService = messenger.messageServiceFromURI(uri);

  try {
    aFileOutputStream.init(tmpFile, -1, parseInt("0600", 8), 0);
  } catch(e) {
debug('aFileOutputStream.init() failed:' + e);
    return;
  }

  var newURI = {};

  let msgWindow = Cc["@mozilla.org/messenger/msgwindow;1"].createInstance(Ci.nsIMsgWindow);
  msgService.CopyMessage(
      uri,
      aCopyListener,
      false,     // aMoveMessage
      null,      // aUrlListener,
      msgWindow, // msgWindow,
      newURI);

//debug('newURI = '+newURI.value.spec);
  newURI = null;
}

function getResentHeaders(msgCompFields, identity)
{
  //the msgCompFields fields are already quoted printable
  //encodeMimeHeader splits them into multiple lines
  let resenthdrs = encodeMimeHeader("Resent-From: " + msgCompFields.from + "\r\n");
  if (msgCompFields.to) {
    resenthdrs += encodeMimeHeader("Resent-To: " + msgCompFields.to + "\r\n");
  }
  if (msgCompFields.cc) {
    resenthdrs += encodeMimeHeader("Resent-CC: " + msgCompFields.cc + "\r\n");
  }
  if (!msgCompFields.to && !msgCompFields.cc) {
    let composeMsgsBundle = Services.strings.createBundle("chrome://messenger/locale/messengercompose/composeMsgs.properties");
    let undisclosedRecipients = composeMsgsBundle.GetStringFromName("undisclosedRecipients");
    resenthdrs += encodeMimeHeader("Resent-To: " + undisclosedRecipients + ":;" + "\r\n");
  }

  resenthdrs += "Resent-Date: " + getResentDate() + "\r\n";
  let msgID = Cc["@mozilla.org/messengercompose/computils;1"].
              createInstance(Ci.nsIMsgCompUtils).
              msgGenerateMessageId(identity);
  if (msgID) {
    resenthdrs += "Resent-Message-ID: " + msgID + "\r\n";
  }

//debug('resent-headers\n'+resenthdrs);
  return resenthdrs;
}


function encodeMimePartIIStr_UTF8(aHeader, aFieldNameLen)
{
    return MailServices.mimeConverter.encodeMimePartIIStr_UTF8(
      aHeader, true, aFieldNameLen, Ci.nsIMimeConverter.MIME_ENCODED_WORD_SIZE);
}
function encodeMimeHeader(header)
{
  let fieldNameLen = (header.indexOf(": ") + 2);
  if (header.length <= MAX_HEADER_LENGTH) {
//header is already quoted printable, encodeMimePartIIStr_UTF8 splits them into multiple lines
    header = header.replace(/\r\n$/, ""); // Don't encode closing end of line
    return header.substr(0, fieldNameLen) + // and don't encode field name
           encodeMimePartIIStr_UTF8(header.substr(fieldNameLen), fieldNameLen) + "\r\n";
  } else {
//header too long, split into multiple headers(!)
    header = header.replace(/\r\n$/, "");
    let fieldName = header.substr(0, fieldNameLen);
    let splitHeader = "";
    let currentLine = "";
    while (header.length > MAX_HEADER_LENGTH - 2) {
      let splitPos = header.substr(0, MAX_HEADER_LENGTH - 2).lastIndexOf(","); // Try to split before MAX_HEADER_LENGTH
      if (splitPos === -1) {
        splitPos = header.indexOf(","); // If that fails, split at first possible position
      }
      if (splitPos === -1) {
        currentLine = header;
        header = "";
      } else {
        currentLine = header.substr(0, splitPos);
        if (header.charAt(splitPos + 1) === " ") {
          header = fieldName + header.substr(splitPos + 2);
        } else {
          header = fieldName + header.substr(splitPos + 1);
        }
      }
      splitHeader += currentLine.substr(0, fieldNameLen) + // Don't encode field name
                     encodeMimePartIIStr_UTF8(currentLine.substr(fieldNameLen), fieldNameLen) + "\r\n";
    }
    splitHeader += header.substr(0, fieldNameLen) + // Don't encode field name
                   encodeMimePartIIStr_UTF8(header.substr(fieldNameLen), fieldNameLen) + "\r\n";
debug('long header: '+splitHeader);
    return(splitHeader);
  }
}


function getResentDate()
{
  let date = new Date();
  let mon=date.toLocaleString('en-US', { month: 'short' });
  let dateTime=date.toLocaleString('en-US', { weekday: 'short' })+', '+
      date.toLocaleString('en-US', { day: 'numeric' })+' '+
      date.toLocaleString('en-US', { month: 'short' })+' '+
      date.toLocaleString('en-US', { year: 'numeric' })+' '+
      date.toLocaleTimeString('de-DE')+' ';
  let offset = date.getTimezoneOffset();
  if (offset < 0) {
    dateTime += "+";
    offset *= -1;
  } else {
    dateTime += "-";
  }
  let minutes = offset % 60;
  offset = (offset - minutes) / 60;
  function twoDigits(aNumber) {
    if (aNumber == 0) {
      return "00";
    }
    return aNumber < 10 ? "0" + aNumber : aNumber;
  }
  dateTime+=twoDigits(offset) + twoDigits(minutes);
debug('resentDate: '+dateTime);
  return dateTime;
}

function stylesheets(context) {
  let styleSheetService = Components.classes["@mozilla.org/content/style-sheet-service;1"]
                                    .getService(Components.interfaces.nsIStyleSheetService);
  let uri = Services.io.newURI(context.extension.getURL("skin/simplemailredirection.css"), null, null);
debug('stylesheet uri='+uri); //fire one time, but
  styleSheetService.loadAndRegisterSheet(uri, styleSheetService.USER_SHEET);
debug('stylesheet loaded');   //this fires two times???
}

let debugcache='';
function debug(txt, force) {
	if (force || prefs) {
		if (force || prefs.debug) {
			if (debugcache) console.log(debugcache); debugcache='';
			console.log('SMR: '+txt);
		}
	} else {
		debugcache+='SMR: '+txt+'\n';
	}
}
